#pragma once

#include "renderer.hpp"

#include "vulkan_allocator.hpp"
#include "vulkan_util.hpp"

namespace gz::vlk {
    struct UniformBufferObject {
        alignas(16) glm::mat4 model;
        alignas(16) glm::mat4 view;
        alignas(16) glm::mat4 projection;
    };

    enum Pipelines3D {
        R3Render,
    };

    class Renderer3D : public Renderer {
        public:
            /**
             * @brief Create a 3D renderer
             * @details
             *  -# @ref VulkanInstance::registerCleanupCallback "register" @ref cleanup() "cleanup callback"
             *  -# @ref VulkanInstance::registerSwapChainRecreateCallback "register" @ref swapChainRecreateCallback() "swapChain recreation callback"
             *  -# create command buffers
             *  -# create vertex & index buffers
             *  -# call createUniformBuffers()
             *  -# call createDescriptorResources()
             *  -# call initSwapChainDependantResources()
             */
            Renderer3D(VulkanInstance& instance, TextureManager& textureManager);
            /**
             * @name Rendering
             */
            /// @{
            void drawFrame(uint32_t imageIndex);
            /// @}
        private:
            void recordCommandBuffer(uint32_t imageIndex, uint32_t currentFrame);
            std::vector<vk::Buffer> uniformBuffers;
            std::vector<MemoryInfo> uniformBuffersMemory;
            void createUniformBuffers();
            void updateUniformBuffer();

            /**
             * @brief Create the render image resources
             * @details
             *  If `RENDERER_3D_RENDER_TO_SWAP_CHAIN`:
             *  - renderImages are null handles
             *  - renderImageMemory are null handles
             *  - renderImageViews are views to VulkanInstance::getScImages()
             *  - renderImageSampler is null handle
             *
             *  Else:
             *  - renderImages are images on renderImageMemory
             *  - renderImageViews are views to renderImages
             *  - renderImageSampler is a combined image sampler
             */
            void createImageResources();

        /**
         * @name Depth
         */
            /// @{
            vk::Image depthImage;
            MemoryInfo depthImageMemory;
            vk::ImageView depthImageView;
            /**
             * @brief Create depth image and view with a @ref VulkanInstance::getDepthFormat() "depth format"
             */
            void createDepthImage();
            /// @}

        /**
         * @name Render pass
         * @details
         *  @section r3p_rp Render pass
         *   @subsection r3p_subpass1 Subpass 1: Rendering
         *    Attributes:
         *    - Pipeline: `Pipelines3D::R3Render`
         *    - Descriptor layouts: `@ref r3p_desc1 "descriptorSet1Layout"` and `@ref tm_desc "TextureManager::descriptorSetLayout"`
         *    - Attachments:
         *      - color blend:
         *        - loadOp = vk::AttachmentLoadOp::eClear
         *        - storeOp = vk::AttachmentStoreOp::eStore
         *        - initialLayout = vk::ImageLayout::eColorAttachmentOptimal
         *        - finalLayout = vk::ImageLayout::eColorAttachmentOptimal
         *        - stencil load/store = dont care
         *   @subsection r3p_subpass2 Subpass 2: Blitting render image onto swap chain image
         *    Attributes:
         *    - Pipeline: `Pipelines3D::blendToSC`
         *    - Descriptor layouts: `@ref r3p_desc2 "descriptorSet2Layout"`
         *    - Attachments:
         *      - color blend:
         *        - loadOp = vk::AttachmentLoadOp::eClear
         *        - storeOp = vk::AttachmentStoreOp::eStore
         *        - initialLayout = vk::ImageLayout::eColorAttachmentOptimal
         *        - finalLayout = vk::ImageLayout::eColorAttachmentOptimal
         *        - stencil load/store = dont care
         */
            /// @{
            /**
             * @brief Create a render pass
             */
            void createRenderPass();
            vk::RenderPass renderPass;
            /// @}

            std::vector<vk::Framebuffer> framebuffers;
            // PIPELINE
            PipelineContainer<Pipelines3D> pipelines;

        /**
         * @name Desciptors
         * @details 
         *  
         *  @section r3d_desc Descriptors
         *   Renderer3D has two descriptor sets, both allocated from the same pool.
         *   @subsection r3d_desc1 Set 1: For rendering 
         *    Bindings:
         *    -0. UniformBufferObject (DESCRIPTOR_TYPE_UNIFORM_BUFFER)
         *   @subsection r3d_desc2 Set 2: For sampling from the rendered image (only if not `RENDERER_3D_RENDER_TO_SWAP_CHAIN)
         *    -0. Image sampler for current render image (DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER)
         */
            /// @{
            vk::DescriptorPool descriptorPool;

            vk::DescriptorSetLayout uboDSL;
            std::vector<vk::DescriptorSet> uboDS;

            vk::DescriptorSetLayout sampleFromRenderImageDSL;
            std::vector<vk::DescriptorSet> sampleFromRenderImageDS;
            /**
             * @brief Create @ref r3d_desc "all descritptor resources"
             */
            void createDescriptorResources();
            /// @}

        /**
         * @name Swapchain recreation
         */
            /// @{
            /**
             * @brief Cleans up resources that were initialized by initSwapChainDependantResources
             */
            void cleanupSwapChainDependantResources();
            /**
             * @brief Sets up resources that depend on the swap chain or its attributes
             * @details
             *  Initializes:
             *  - @ref createRenderPass() "render pass"
             *  - @ref createImages() "images and imageViews"
             *  - @ref createDepthImage() "depth image and view"
             *  - @ref VulkanInstance::createFramebuffers() "framebuffers"
             *  - pipelines:
             *    - R3Render
             */
            void initSwapChainDependantResources();
            /**
             * @brief Recreates swap chain dependant resources
             * @details
             *  Calls:
             *  -# cleanupSwapChainDependantResources()
             *  -# initSwapChainDependantResources()
             */
            void swapChainRecreateCallback();
            /// @}

            /*
             * @brief Destroy all vulkan objects owned by this object
             * @details:
             *  Does:
             *  -# destroy uniform buffers
             *  -# destroy descriptor set layout and pool
             *  -# call cleanupSwapChainDependantResources()
             *  -# call Renderer::cleanup_()
             */
            void cleanup();

            void loadModel();
            VerticesAndIndices<uint32_t> model;

            Log rLog;
    };
}
